from ..message_pb2 import Message


def decode(data: bytes) -> Message:
    """
    Decode the given data using the Message.FromString method.

    Args:
        data (bytes): The data to be decoded.

    Returns:
        Message: The decoded message object.
    """
    return Message.FromString(data)


def encode(data: Message) -> bytes:
    """
    Encodes the given data object into a serialized byte string.

    Args:
        data: The data object to be encoded.

    Returns:
        str: The serialized string representation of the data object.
    """
    return data.SerializeToString()



import math
import uuid

from .message_pb2 import Message
from .mutex import mutex
from .topology import Topology


class Gossip:
    FANOUT = 5

    def __init__(self, transport, fanout=None, peer_id=None):
        self.peer_id = peer_id or uuid.uuid4().bytes
        self.transport = transport
        self.topology = Topology(self.peer_id, self.transport.addr)

        self._fanout = fanout or self.FANOUT

    @property
    def fanout(self):
        return min(self._fanout, len(self.topology))

    @property
    def cycles(self):
        if self.fanout == 0:
            return 0

        if self.fanout == 1:
            return 1

        return math.ceil(math.log(len(self.topology), self.fanout))

    async def send(self, message, peer_id):
        if not message.id:
            raise ValueError("message id is required:", message)

        if not message.kind:
            raise ValueError("message kind is required:", message)

        if not message.routing.src_id:
            raise ValueError("message routing.src_id is required:", message)

        if not message.routing.dst_id:
            raise ValueError("message routing.dst_id is required:", message)

        msg = Message()
        msg.CopyFrom(message)
        msg = self.topology.append_route(msg)

        addr = self.topology.get_addr(peer_id)
        await self.transport.send(msg, addr)
        return msg.id

    async def send_handshake(self, peer_id):
        msg = Message()
        msg.id = uuid.uuid4().bytes
        msg.kind.append(Message.Kind.HANDSHAKE)
        msg.kind.append(Message.Kind.SYN)
        msg.routing.src_id = self.peer_id
        msg.routing.dst_id = peer_id

        return await self.send(msg, peer_id)

    async def send_forward(self, message):
        msg = Message()
        msg.CopyFrom(message)

        return await self.send(msg, message.routing.dst_id)

    async def send_ack(self, message):
        if message.Kind.SYN not in message.kind:
            raise ValueError("SYN message is required for ACK:", message)

        msg = Message()
        msg.id = message.id
        msg.kind.append(Message.Kind.ACK)
        msg.routing.src_id = self.peer_id
        msg.routing.dst_id = message.routing.src_id
        msg.topic = message.topic

        return await self.send(msg, message.routing.src_id)

    async def send_gossip(self, message):
        msg = Message()
        msg.CopyFrom(message)

        if not msg.id:
            msg.id = uuid.uuid4().bytes

        if Message.Kind.GOSSIP not in msg.kind:
            msg.kind.append(Message.Kind.GOSSIP)

        if not msg.routing.src_id:
            msg.routing.src_id = self.peer_id

        messages = set()
        gossip_ignore = set([self.peer_id])
        gossip_ignore.update([r.route_id for r in msg.routing.routes])

        @mutex(self, msg.id)
        async def multicast():
            cycle = 0
            while cycle < self.cycles:
                peer_ids = self.topology.sample(self.fanout, ignore=gossip_ignore)
                for peer_id in peer_ids:
                    m = Message()
                    m.CopyFrom(msg)
                    m.routing.dst_id = peer_id
                    messages.add(await self.send(m, peer_id))
                gossip_ignore.update(peer_ids)
                cycle += 1

        await multicast()
        return list(messages)

    async def send_gossip_handshake(self):
        msg = Message()
        msg.id = uuid.uuid4().bytes
        msg.kind.append(Message.Kind.HANDSHAKE)
        msg.kind.append(Message.Kind.SYN)
        msg.routing.src_id = self.peer_id

        return await self.send_gossip(msg)

    async def recv(self):
        while True:
            msg, addr = await self.transport.recv()
            msg.routing.routes[-1].daddr = f"{addr[0]}:{addr[1]}"

            msg = self.topology.append_route(msg)
            route_ids = self.topology.update(msg.routing.routes)
            for route_id in route_ids:
                await self.send_handshake(route_id)

            # forward message to destination
            if self.peer_id != msg.routing.dst_id:
                await self.send_forward(msg)
                continue

            # ack message
            if Message.Kind.ACK in msg.kind:
                yield msg
                continue

            # syn message
            if Message.Kind.SYN in msg.kind:
                await self.send_ack(msg)

            # gossip message
            if Message.Kind.GOSSIP in msg.kind:
                await self.send_gossip(msg)

            # handshake message
            if Message.Kind.HANDSHAKE in msg.kind:
                continue

            yield msg

    async def close(self):
        self.transport.close()



import collections
import functools
import time

from .config import MUTEX_TTL

MUTEX = collections.defaultdict(dict)


def mutex(owner, mutex_id, mutex_ttl=MUTEX_TTL):
    def decorator(func):
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            mut = MUTEX[owner].get(mutex_id)
            if mut:
                if time.time() - mut > mutex_ttl:
                    MUTEX[owner][mutex_id] = time.time()
                    return await func(*args, **kwargs)
                else:
                    return
            else:
                MUTEX[owner][mutex_id] = time.time()
                return await func(*args, **kwargs)

        return wrapper

    return decorator



"""
Usage:
    GOSSIP_SEEDS=response@127.0.0.1:1337 python examples/reqresp/req.py
"""
import aiogossip

peer = aiogossip.Peer(peer_id="request")


async def main():
    message = aiogossip.Message()

    response = await peer.request("query", message)
    async for resp in response:
        print(resp)


if __name__ == "__main__":
    print("DSN:", peer.DSN)
    peer.connect()
    peer.run_forever(main=main)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: message.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rmessage.proto\"J\n\x05Route\x12\x10\n\x08route_id\x18\x01 \x01(\x0c\x12\x11\n\ttimestamp\x18\x02 \x01(\x03\x12\r\n\x05saddr\x18\x03 \x01(\t\x12\r\n\x05\x64\x61\x64\x64r\x18\x04 \x01(\t\"\x91\x02\n\x07Message\x12\n\n\x02id\x18\x01 \x01(\x0c\x12\x1b\n\x04kind\x18\x02 \x03(\x0e\x32\r.Message.Kind\x12!\n\x07routing\x18\x03 \x01(\x0b\x32\x10.Message.Routing\x12\r\n\x05topic\x18\x04 \x01(\t\x12\x0f\n\x07payload\x18\x05 \x01(\x0c\x1a\x41\n\x07Routing\x12\x0e\n\x06src_id\x18\x01 \x01(\x0c\x12\x0e\n\x06\x64st_id\x18\x02 \x01(\x0c\x12\x16\n\x06routes\x18\x03 \x03(\x0b\x32\x06.Route\"W\n\x04Kind\x12\r\n\tHANDSHAKE\x10\x00\x12\n\n\x06GOSSIP\x10\x01\x12\x07\n\x03SYN\x10\x02\x12\x07\n\x03\x41\x43K\x10\x03\x12\x07\n\x03PUB\x10\x04\x12\x07\n\x03SUB\x10\x05\x12\x07\n\x03REQ\x10\x06\x12\x07\n\x03RES\x10\x07\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'message_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_ROUTE']._serialized_start=17
  _globals['_ROUTE']._serialized_end=91
  _globals['_MESSAGE']._serialized_start=94
  _globals['_MESSAGE']._serialized_end=367
  _globals['_MESSAGE_ROUTING']._serialized_start=213
  _globals['_MESSAGE_ROUTING']._serialized_end=278
  _globals['_MESSAGE_KIND']._serialized_start=280
  _globals['_MESSAGE_KIND']._serialized_end=367
# @@protoc_insertion_point(module_scope)



import asyncio
import collections
import fnmatch
import itertools

from .channel import Channel
from .gossip import Gossip
from .message_pb2 import Message
from .task_manager import TaskManager


class Handler:
    def __init__(self, topic, func, loop: asyncio.AbstractEventLoop):
        self._loop = loop

        self.topic = topic
        self.func = func

        self.chan = Channel(loop=self._loop)
        self.task_manager = TaskManager(loop=self._loop)
        self.task_manager.create_task(self())

        self.hooks = []

    async def __call__(self):
        while True:
            message = await self.chan.recv()
            result = await self.func(message)

            if result is None:
                continue

            for hook in self.hooks:
                self.task_manager.create_task(hook(message, result))

    def hook(self, func):
        self.hooks.append(func)

    async def cancel(self):
        await self.chan.close()
        await self.task_manager.close()


class Broker:
    TIMEOUT = 1

    def __init__(self, gossip: Gossip, loop: asyncio.AbstractEventLoop = None):
        self._loop = loop or gossip.transport._loop

        self.gossip = gossip
        self.handlers = collections.defaultdict(list)

    async def listen(self):
        """Connect to the gossip network and start receiving messages."""
        async for message in self.gossip.recv():
            # FIXME: make messages idempotent (prevent duplicate processing)

            handlers = list(self.handlers.keys())

            if f"recv:{message.id}" in handlers:
                for handler in self.handlers[f"recv:{message.id}"]:
                    await handler.chan.send(message)

            elif message.topic:
                for topic in handlers:
                    if fnmatch.fnmatch(message.topic, topic):
                        for handler in self.handlers[topic]:
                            await handler.chan.send(message)

            else:
                continue

            for topic in handlers:
                if len(self.handlers[topic]) == 0:
                    del self.handlers[topic]

    async def close(self):
        """Disconnect from the gossip network and stop receiving messages."""
        handlers = [h for h in itertools.chain(*self.handlers.values())]
        await asyncio.gather(*[h.cancel() for h in handlers], return_exceptions=True)
        await self.gossip.close()

    def subscribe(self, topic, func):
        """Subscribe to a topic and register a handler."""
        handler = Handler(topic, func, loop=self._loop)
        self.handlers[topic].append(handler)
        return handler

    async def unsubscribe(self, handler):
        """Unsubscribe from a topic and unregister a handler."""
        await handler.cancel()
        self.handlers[handler.topic].remove(handler)

    async def publish(self, topic, message, peer_ids=None):
        """Publish a message to a topic."""
        # FIXME: make messages idempotent (prevent duplicate processing)
        # FIXME: allow sending to specific nodes
        message.topic = topic
        if not message.id:
            raise ValueError("Message ID is required:", message)
        message_id = message.id

        if peer_ids:
            for peer_id in peer_ids:
                if peer_id in self.gossip.topology:
                    await self.gossip.send(message, peer_id)
                else:
                    raise ValueError(f"Unknown node: {peer_id}")
        else:
            await self.gossip.send_gossip(message)

        return self._recv(message_id, peer_ids=peer_ids)

    async def _recv(self, message_id, peer_ids=None):
        chan = Channel(loop=self._loop)
        handler = self.subscribe(f"recv:{message_id}", chan.send)

        acks = set()
        if peer_ids:
            acks.update(peer_ids)

        try:
            async with asyncio.timeout(self.TIMEOUT):
                while True:
                    message = await chan.recv()

                    if Message.Kind.ACK in message.kind:
                        acks.add(message.routing.src_id)
                        yield message  # FIXME: don't yield acks

                    elif message.routing.src_id in acks:
                        acks.remove(message.routing.src_id)
                        yield message

                    else:
                        raise ValueError(f"Unknown message: {message}")

        except asyncio.TimeoutError:
            pass
        finally:
            await self.unsubscribe(handler)
            await chan.close()



"""
Usage:
    GOSSIP_ID=p1 GOSSIP_PORT=1337 python examples/members/members.py
    GOSSIP_ID=p2 GOSSIP_SEEDS=p1@127.0.0.1:1337 python examples/members/members.py
"""
import asyncio
import os

import aiogossip

peer_id = os.getenv("GOSSIP_ID")
port = int(os.getenv("GOSSIP_PORT", 0))
peer = aiogossip.Peer(port=port, peer_id=peer_id)


async def main():
    while True:
        await asyncio.sleep(10)
        print("---")
        print("members:")
        for node in peer.nodes:
            reachable = peer.gossip.topology[node]["reachable"]
            print("-", node, "reachable" if reachable else "unreachable")
        print()


if __name__ == "__main__":
    print("DSN:", peer.DSN)
    peer.connect()
    peer.run_forever(main=main)



import asyncio
import collections


class Channel:
    """A simple asynchronous channel for sending and receiving messages.

    This class represents a channel that allows sending and receiving messages asynchronously.
    Messages can be sent using the `send` method, and received using the `recv` method.
    If the channel is empty, the `recv` method will wait until a message is sent.

    """

    def __init__(self, loop: asyncio.AbstractEventLoop):
        """Initialize the channel with empty queue and waiters.

        Args:
            loop (asyncio.AbstractEventLoop): The event loop to associate with the channel.

        """
        self._loop = loop

        self._queue = collections.deque()
        self._waiters = collections.deque()

    async def send(self, message):
        """Send a message to the channel. If there are any waiters, wake one up.

        Args:
            message: The message to send.

        """
        self._queue.append(message)
        if self._waiters:
            waiter = self._waiters.popleft()
            if not waiter.done():
                waiter.set_result(None)

    async def recv(self):
        """Receive a message from the channel. If the channel is empty, wait until a message is sent.

        Returns:
            The received message.

        """
        while not self._queue:
            waiter = self._loop.create_future()
            self._waiters.append(waiter)
            try:
                await waiter
            except asyncio.CancelledError:
                if waiter in self._waiters:
                    self._waiters.remove(waiter)
                raise
        return self._queue.popleft()

    async def close(self):
        """Close the channel. Any pending waiters are cancelled."""
        for waiter in self._waiters:
            waiter.cancel()
        self._waiters.clear()
        self._queue.clear()



import asyncio
import logging
import sys

from . import config
from .message_pb2 import Message
from .task_manager import TaskManager

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler(sys.stdout))
if config.DEBUG:
    logger.setLevel(logging.DEBUG)  # pragma: no cover


class Members:
    SCHEDULER_INTERVAL = 5

    PING_INTERVAL = 15
    PING_TIMEOUT = 5

    def __init__(self, peer):
        self.peer = peer

        self.task_manager = TaskManager()
        self.task_manager.create_task(self.scheduler())

        self.peer.response("keepalive:*")(self.pong)

    async def scheduler(self):
        while True:
            for node in self.peer.nodes:
                if node == self.peer.peer_id:
                    continue

                if node not in self.task_manager:
                    self.task_manager.create_task(self.ping(node))

            await asyncio.sleep(self.SCHEDULER_INTERVAL)

    async def ping(self, peer_id):
        while True:
            topic = "keepalive"
            message = Message()
            message.routing.src_id = self.peer.peer_id
            message.routing.dst_id = peer_id

            response = await self.peer.request(topic, message, peers=[peer_id], timeout=self.PING_TIMEOUT)
            responses = []
            async for r in response:
                responses.append(r)

            if len(responses):
                self.peer.gossip.topology.mark_reachable(peer_id)
                logger.debug(f"Node is reachable: {peer_id}")
            else:
                self.peer.gossip.topology.mark_unreachable(peer_id)
                logger.debug(f"Node is unreachable: {peer_id}")

            await asyncio.sleep(self.PING_INTERVAL)

    async def pong(self, message):
        return Message()

    async def close(self):
        await self.task_manager.close()



import pdb
import sys

from . import config


def debug(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            if config.DEBUG and sys.stdout.isatty():
                pdb.post_mortem()  # pragma: no cover
            raise

    return wrapper



"""
Usage:
    python examples/reqresp/resp.py
"""
import aiogossip

peer = aiogossip.Peer(port=1337, peer_id="response")


@peer.response("query")
async def query(message):
    return aiogossip.Message(payload=b"response")


if __name__ == "__main__":
    print("DSN:", peer.DSN)
    peer.connect()
    peer.run_forever()



import asyncio
import uuid

from . import config
from .broker import Broker
from .debug import debug
from .gossip import Gossip
from .members import Members
from .message_pb2 import Message
from .task_manager import TaskManager
from .topology import Node
from .transport import Transport
from .transport.address import parse_addr


class Peer:
    def __init__(
        self,
        host="0.0.0.0",
        port=0,
        fanout=None,
        peer_id=None,
        loop: asyncio.AbstractEventLoop = None,
    ):
        self._loop = loop or asyncio.get_event_loop()

        if peer_id:
            self.peer_id = peer_id.encode()
        else:
            self.peer_id = uuid.uuid1().bytes

        self.transport = Transport((host, port), loop=self._loop)
        self.gossip = Gossip(self.transport, fanout=fanout, peer_id=self.peer_id)
        self.broker = Broker(self.gossip, loop=self._loop)

        self.task_manager = TaskManager(loop=self._loop)
        self.task_manager.create_task(self.broker.listen())

        self.members = Members(self)

    @property
    def node(self):
        return self.gossip.topology.node

    @property
    def nodes(self):
        return list(self.gossip.topology)

    @property
    def DSN(self):
        return "{}@{}:{}".format(self.node.node_id.decode(), *self.node.node_addr)

    async def _connect(self):
        await self.gossip.send_gossip_handshake()

    def connect(self, seeds=config.SEEDS):
        if not seeds:
            return

        nodes = []

        if isinstance(seeds, str):
            seeds = seeds.split(",")
            for seed in seeds:
                peer_id, addr = seed.split("@")
                addr = parse_addr(addr)
                nodes.append(Node(node_id=peer_id.encode(), node_addr=addr))

        elif isinstance(seeds, list):
            nodes = seeds

        self.gossip.topology.add(nodes)
        self.task_manager.create_task(self._connect())

    async def disconnect(self):
        await self.members.close()
        await self.broker.close()
        await self.task_manager.close()

    async def publish(self, topic, message, peers=None, syn=False):
        if not message.id:
            message.id = uuid.uuid4().bytes
        topic = topic.replace("{uuid}", uuid.uuid4().hex)

        if syn and (Message.Kind.SYN not in message.kind):
            message.kind.append(Message.Kind.SYN)
        return await self.broker.publish(topic, message, peer_ids=peers)

    def subscribe(self, topic):
        def decorator(func):
            return self.broker.subscribe(topic, func)

        return decorator

    async def request(self, topic, message, peers=None, timeout=5):
        if not message.id:
            message.id = uuid.uuid4().bytes
        message.kind.append(Message.Kind.REQ)

        topic = "request:{}:{}".format(topic, uuid.uuid4().hex)
        response = await self.publish(topic, message, peers=peers, syn=True)
        return response

    def response(self, topic):
        topic = "request:{}:*".format(topic)

        async def responder(message, result):
            result.id = message.id
            result.routing.src_id = message.routing.dst_id
            result.routing.dst_id = message.routing.src_id
            result.kind.append(Message.Kind.RES)
            await self.publish(message.topic, result, peers=[message.routing.src_id], syn=False)

        def decorator(func):
            handler = self.broker.subscribe(topic, func)
            handler.hook(responder)
            return handler

        return decorator

    @debug
    def run_forever(self, main=None):  # pragma: no cover
        try:
            if main:
                main = self._loop.create_task(main())
            self._loop.run_forever()
        except KeyboardInterrupt:
            pass
        finally:
            if main:
                main.cancel()
                self._loop.run_until_complete(main)
            self._loop.run_until_complete(self.disconnect())



import collections
import ipaddress
import warnings

Address = collections.namedtuple("Address", ["ip", "port"])


def parse_addr(addr: Address | tuple | str) -> Address:
    """
    Parse the given address and return an Address object.

    Args:
        addr (Address, tuple, str): The address to parse. It can be an Address object,
            a tuple containing the IP address and port, or a string in the format "ip:port".

    Returns:
        Address: The parsed Address object.

    Raises:
        TypeError: If the address is not of type Address, tuple, or str.

    Deprecated:
        Passing a tuple or string to parse_addr is deprecated. Use Address instead.
    """
    if isinstance(addr, Address):
        return addr
    elif isinstance(addr, tuple):
        warnings.warn(
            "Passing a tuple to parse_addr is deprecated, use Address instead",
            DeprecationWarning,
        )
        return Address(ipaddress.ip_address(addr[0]), int(addr[1]))
    elif isinstance(addr, str):
        warnings.warn(
            "Passing a string to parse_addr is deprecated, use Address instead",
            DeprecationWarning,
        )
        ip, port = addr.split(":")
        return Address(ipaddress.ip_address(ip), int(port))
    else:
        raise TypeError(f"Address must be a Address, tuple or str, got: {type(addr)}")



class Gossip:
    def __init__(self, transport, fanout=None, peer_id=None): ...
    async def send(self, message, peer_id): ...
    async def send_ack(self, message): ...
    async def send_forward(self, message): ...
    async def send_gossip(self, message): ...
    async def recv(self): ...
    async def close(self): ...



import asyncio


class TaskManager:
    def __init__(self, loop=None):
        self.loop = loop or asyncio.get_event_loop()

        self.tasks = []
        self.named_tasks = {}

    def create_task(self, coro, name=None):
        task = self.loop.create_task(coro)
        task.add_done_callback(self.on_task_done)
        self.tasks.append(task)
        if name:
            self.named_tasks[name] = task
        return task

    def on_task_done(self, task):
        self.tasks.remove(task)
        for name, t in self.named_tasks.items():
            if t == task:
                del self.named_tasks[name]
                break

        try:
            exception = task.exception()
        except (asyncio.CancelledError, asyncio.InvalidStateError):
            return

        if not exception:
            return

        task.print_stack()

    def cancel(self):
        for task in self.tasks:
            task.cancel()
        self.tasks = []
        self.named_tasks = {}

    async def close(self):
        self.cancel()
        await asyncio.gather(*self.tasks, return_exceptions=True)

    def __getitem__(self, item):
        return self.named_tasks[item]



import asyncio
import ipaddress
import logging
import os
import socket
import sys

from . import codec
from .address import Address

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler(sys.stdout))
if os.environ.get("DEBUG"):
    logger.setLevel(logging.DEBUG)  # pragma: no cover


class Transport:
    """
    Represents a transport layer for sending and receiving messages over UDP.
    """

    PACKET_SIZE = 4096

    def __init__(self, bind, loop: asyncio.AbstractEventLoop):
        self._loop = loop

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.bind(bind)
        self.sock.setblocking(False)

        self.rx_bytes = 0
        self.rx_packets = 0
        self.tx_bytes = 0
        self.tx_packets = 0

    @property
    def addr(self):
        """
        Returns the local address of the transport.
        """
        ip, port = self.sock.getsockname()
        ip = ipaddress.ip_address(ip)
        return Address(ip, port)

    async def send(self, message, addr: Address):
        """
        Sends a message to the specified address.

        Args:
            message: The message to send.
            addr: The address to send the message to.

        Raises:
            TypeError: If the address is not of type Address.
            ValueError: If the message size exceeds the packet size.

        Returns:
            None
        """
        if not isinstance(addr, Address):
            raise TypeError(f"Address must be a tuple, got: {type(addr)}")
        if not isinstance(addr.ip, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
            raise TypeError(f"IP address must be IPv4, got: {type(addr.ip)}")
        if not isinstance(addr.port, int):
            raise TypeError(f"Port must be an integer, got: {type(addr.port)}")

        msg = codec.encode(message)
        if len(msg) > self.PACKET_SIZE:
            raise ValueError(f"Message size exceeds packet size of {self.PACKET_SIZE} bytes: {len(msg)}")

        await self._loop.sock_sendto(self.sock, msg, (addr.ip.exploded, addr.port))
        self.tx_packets += 1
        self.tx_bytes += len(msg)

        logger.debug(f"DEBUG: {self.addr[1]} > {addr[1]} send: {message}\n")

    async def recv(self):
        """
        Receives a message from the transport.

        Returns:
            A tuple containing the received message and the address it was received from.
        """
        msg, addr = await self._loop.sock_recvfrom(self.sock, self.PACKET_SIZE)
        self.rx_packets += 1
        self.rx_bytes += len(msg)

        addr = Address(ipaddress.ip_address(addr[0]), int(addr[1]))
        message = codec.decode(msg)
        logger.debug(f"DEBUG: {self.addr.port} < {addr.port} recv: {message}\n")

        return message, addr

    def close(self):
        """
        Closes the transport.
        """
        self.sock.close()



"""
Usage:
    GOSSIP_SEEDS=subscriber@127.0.0.1:1337 python examples/pubsub/pub.py
"""
import aiogossip

peer = aiogossip.Peer(peer_id="publisher")


async def main():
    message = aiogossip.Message(payload=b"payload")
    await peer.publish("topic", message)


if __name__ == "__main__":
    print("DSN:", peer.DSN)
    peer.connect()
    peer.run_forever(main=main)



import collections
import random
import time

import networkx as nx

from .message_pb2 import Message, Route
from .transport.address import parse_addr

Node = collections.namedtuple("Node", ["node_id", "node_addr"])


class Topology:
    def __init__(self, node_id, node_addr):
        self.g = nx.DiGraph(node_id=node_id, node_addr=node_addr)
        self.g.add_node(
            node_id,
            node_id=node_id,
            node_addr=parse_addr(node_addr),
            reachable=True,
        )

    # Node #
    @property
    def node_id(self):
        return self.g.graph["node_id"]

    @property
    def node_addr(self):
        return self.g.graph["node_addr"]

    @property
    def node(self):
        return Node(self.node_id, self.node_addr)

    # Topology #
    def add(self, nodes):
        for node in nodes:
            if node.node_id == self.node_id:
                continue

            self.g.add_node(
                node.node_id,
                node_id=node.node_id,
                node_addr=parse_addr(node.node_addr),
            )

            src = self.g.nodes[self.node_id]
            dst = self.g.nodes[node.node_id]
            self.g.add_edge(
                src["node_id"],
                dst["node_id"],
                saddr=parse_addr(src["node_addr"]),
                daddr=parse_addr(dst["node_addr"]),
            )

    def update(self, routes):
        if len(routes) < 2:
            raise ValueError("Empty route")

        nodes = set()
        for r in (r for r in routes if r.route_id not in self.g):
            self.g.add_node(r.route_id, node_id=r.route_id, node_addr=parse_addr(r.daddr or r.saddr))
            nodes.add(r.route_id)

        def edge(src, dst):
            return {
                "saddr": parse_addr(src.daddr or src.saddr),
                "daddr": parse_addr(dst.daddr or dst.saddr),
                "latency": abs(src.timestamp - dst.timestamp),
            }

        hops = ((routes[r], routes[r + 1]) for r in range(len(routes) - 1))
        for src, dst in hops:
            self.g.add_edge(src.route_id, dst.route_id, **edge(src, dst))
        self.g.add_edge(dst.route_id, src.route_id, **edge(dst, src))

        return nodes

    def sample(self, k, ignore=None):
        nodes = [e[1] for e in self.g.out_edges(self.node_id)]
        if ignore:
            nodes = list(set(nodes) - set(ignore))
        k = min(k, len(nodes))
        return random.sample(nodes, k)

    def __getitem__(self, node_id):
        return self.g.nodes[node_id]

    def __len__(self):
        return len(self.g)

    def __iter__(self):
        return iter(self.g.nodes)

    def __contains__(self, node_id):
        return node_id in self.g

    # Reachability #

    def mark_reachable(self, node_id):
        self.g.nodes[node_id]["reachable"] = True

    def mark_unreachable(self, node_id):
        self.g.nodes[node_id]["reachable"] = False

    # Route #
    @property
    def route(self):
        return Route(
            route_id=self.node_id,
            timestamp=int(time.time_ns()),
            saddr=f"{self.node.node_addr.ip}:{self.node.node_addr.port}",
        )

    def append_route(self, message):
        msg = Message()
        msg.CopyFrom(message)

        if not msg.routing.routes:
            msg.routing.routes.append(self.route)

        if msg.routing.routes[-1].route_id != self.node_id:
            msg.routing.routes.append(self.route)

        return msg

    # Addr #
    def get_addr(self, node_id):
        path = nx.shortest_path(self.g.to_undirected(), self.node_id, node_id)
        addr = self.g.edges[path[0], path[1]]["daddr"]
        return parse_addr(addr)



"""
Usage:
    python examples/pubsub/sub.py
"""
import aiogossip

peer = aiogossip.Peer(port=1337, peer_id="subscriber")


@peer.subscribe("*")
async def on_message(message):
    print(message)


if __name__ == "__main__":
    print("DSN:", peer.DSN)
    peer.connect()
    peer.run_forever()
